/*
 * Copyright (C) NHR@FAU, University Erlangen-Nuremberg.
 * All rights reserved. This file is part of nusif-solver.
 * Use of this source code is governed by a MIT style
 * license that can be found in the LICENSE file.
 */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#include "comm.h"

#ifdef _MPI
// subroutines local to this module
static int sum(int* sizes, int position)
{
    int sum = 0;

    for (int i = 0; i < position; i++) {
        sum += sizes[i];
    }

    return sum;
}

static void assembleResult(Comm* c, double* src, double* dst, int jmax, int imax)
{
    MPI_Request* requests;
    int numRequests = 1;

    if (c->rank == 0) {
        numRequests = c->size + 1;
    } else {
        numRequests = 1;
    }

    requests = (MPI_Request*)malloc(numRequests * sizeof(MPI_Request));

    /* all ranks send their bulk array, but including external boundary layer */
    MPI_Datatype bulkType;
    int oldSizes[NDIMS] = { c->jmaxLocal + 2, c->imaxLocal + 2 };
    int newSizes[NDIMS] = { c->jmaxLocal, c->imaxLocal };
    int starts[NDIMS]   = { 1, 1 };

    if (commIsBoundary(c, LEFT)) {
        newSizes[IDIM] += 1;
        starts[IDIM] = 0;
    }
    if (commIsBoundary(c, RIGHT)) {
        newSizes[IDIM] += 1;
    }
    if (commIsBoundary(c, BOTTOM)) {
        newSizes[JDIM] += 1;
        starts[JDIM] = 0;
    }
    if (commIsBoundary(c, TOP)) {
        newSizes[JDIM] += 1;
    }

    MPI_Type_create_subarray(NDIMS,
        oldSizes,
        newSizes,
        starts,
        MPI_ORDER_C,
        MPI_DOUBLE,
        &bulkType);
    MPI_Type_commit(&bulkType);
    MPI_Isend(src, 1, bulkType, 0, 0, c->comm, &requests[0]);

    int newSizesI[c->size];
    int newSizesJ[c->size];
    MPI_Gather(&newSizes[IDIM], 1, MPI_INT, newSizesI, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Gather(&newSizes[JDIM], 1, MPI_INT, newSizesJ, 1, MPI_INT, 0, MPI_COMM_WORLD);

    /* rank 0 assembles the subdomains */
    if (c->rank == 0) {
        for (int i = 0; i < c->size; i++) {
            MPI_Datatype domainType;
            int oldSizes[NDIMS] = { jmax + 2, imax + 2 };
            int newSizes[NDIMS] = { newSizesJ[i], newSizesI[i] };
            int coords[NDIMS];
            MPI_Cart_coords(c->comm, i, NDIMS, coords);
            int starts[NDIMS] = { sum(newSizesJ, coords[JDIM]),
                sum(newSizesI, coords[IDIM]) };
            MPI_Type_create_subarray(NDIMS,
                oldSizes,
                newSizes,
                starts,
                MPI_ORDER_C,
                MPI_DOUBLE,
                &domainType);
            MPI_Type_commit(&domainType);

            MPI_Irecv(dst, 1, domainType, i, 0, c->comm, &requests[i + 1]);
        }
    }

    MPI_Waitall(numRequests, requests, MPI_STATUSES_IGNORE);
}
#endif // defined _MPI

// exported subroutines
int commIsBoundary(Comm* c, int direction)
{
#ifdef _MPI
    switch (direction) {
    case LEFT:
        return c->coords[IDIM] == 0;
        break;
    case RIGHT:
        return c->coords[IDIM] == (c->dims[IDIM] - 1);
        break;
    case BOTTOM:
        return c->coords[JDIM] == 0;
        break;
    case TOP:
        return c->coords[JDIM] == (c->dims[JDIM] - 1);
        break;
    }
#endif

    return 1;
}

void commExchange(Comm* c, double* grid)
{
#ifdef _MPI
    int counts[NDIRS] = { 1, 1, 1, 1 };

    MPI_Neighbor_alltoallw(grid,
        counts,
        c->sdispls,
        c->bufferTypes,
        grid,
        counts,
        c->rdispls,
        c->bufferTypes,
        c->comm);
#endif
}

void commShift(Comm* c, double* f, double* g)
{
#ifdef _MPI
    MPI_Request requests[4] = { MPI_REQUEST_NULL,
        MPI_REQUEST_NULL,
        MPI_REQUEST_NULL,
        MPI_REQUEST_NULL };

    /* shift G */
    /* receive ghost cells from bottom neighbor */
    double* buf = g + 1;
    MPI_Irecv(buf,
        1,
        c->bufferTypes[BOTTOM],
        c->neighbours[BOTTOM],
        0,
        c->comm,
        &requests[0]);

    /* send ghost cells to top neighbor */
    buf = g + (c->jmaxLocal) * (c->imaxLocal + 2) + 1;
    MPI_Isend(buf, 1, c->bufferTypes[TOP], c->neighbours[TOP], 0, c->comm, &requests[1]);

    /* shift F */
    /* receive ghost cells from left neighbor */
    buf = f + (c->imaxLocal + 2);
    MPI_Irecv(buf,
        1,
        c->bufferTypes[LEFT],
        c->neighbours[LEFT],
        1,
        c->comm,
        &requests[2]);

    /* send ghost cells to right neighbor */
    buf = f + (c->imaxLocal + 2) + (c->imaxLocal);
    MPI_Isend(buf,
        1,
        c->bufferTypes[RIGHT],
        c->neighbours[RIGHT],
        1,
        c->comm,
        &requests[3]);

    MPI_Waitall(4, requests, MPI_STATUSES_IGNORE);
#endif
}

void commCollectResult(Comm* c,
    double* ug,
    double* vg,
    double* pg,
    double* u,
    double* v,
    double* p,
    int jmax,
    int imax)
{
#ifdef _MPI

    int offset[c->size * NDIMS];
    int imaxLocal[c->size];
    int jmaxLocal[c->size];

    MPI_Gather(&c->imaxLocal, 1, MPI_INT, imaxLocal, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Gather(&c->jmaxLocal, 1, MPI_INT, jmaxLocal, 1, MPI_INT, 0, MPI_COMM_WORLD);

    if (c->rank == 0) {
        for (int i = 0; i < c->size; i++) {
            int coords[NDIMS];
            MPI_Cart_coords(c->comm, i, NDIMS, coords);
            offset[i * NDIMS + IDIM] = sum(imaxLocal, coords[IDIM]);
            offset[i * NDIMS + JDIM] = sum(jmaxLocal, coords[JDIM]);
            printf("Rank: %d, Coords(j,i): %d %d, Size(j,i): %d %d "
                   "Offset(j,i): %d %d\n",
                i,
                coords[JDIM],
                coords[IDIM],
                jmaxLocal[i],
                imaxLocal[i],
                offset[i * NDIMS + JDIM],
                offset[i * NDIMS + IDIM]);
        }
    }

    /* collect P */
    assembleResult(c, p, pg, jmax, imax);

    /* collect U */
    assembleResult(c, u, ug, jmax, imax);

    /* collect V */
    assembleResult(c, v, vg, jmax, imax);
#endif
}

void commPartition(Comm* c, int jmax, int imax)
{
#ifdef _MPI
    int dims[NDIMS]    = { 0, 0 };
    int periods[NDIMS] = { 0, 0 };
    MPI_Dims_create(c->size, NDIMS, dims);
    MPI_Cart_create(MPI_COMM_WORLD, NDIMS, dims, periods, 0, &c->comm);
    MPI_Cart_shift(c->comm, IDIM, 1, &c->neighbours[LEFT], &c->neighbours[RIGHT]);
    MPI_Cart_shift(c->comm, JDIM, 1, &c->neighbours[BOTTOM], &c->neighbours[TOP]);
    MPI_Cart_get(c->comm, NDIMS, c->dims, periods, c->coords);

    c->imaxLocal = sizeOfRank(c->rank, dims[IDIM], imax);
    c->jmaxLocal = sizeOfRank(c->rank, dims[JDIM], jmax);

    MPI_Datatype jBufferType;
    MPI_Type_contiguous(c->imaxLocal, MPI_DOUBLE, &jBufferType);
    MPI_Type_commit(&jBufferType);

    MPI_Datatype iBufferType;
    MPI_Type_vector(c->jmaxLocal, 1, c->imaxLocal + 2, MPI_DOUBLE, &iBufferType);
    MPI_Type_commit(&iBufferType);

    // in the order of the dimensions i->0, j->1
    // first negative direction, then positive direction
    size_t dblsize         = sizeof(double);
    int imaxLocal          = c->imaxLocal;
    int jmaxLocal          = c->jmaxLocal;
    c->bufferTypes[LEFT]   = iBufferType;
    c->bufferTypes[RIGHT]  = iBufferType;
    c->bufferTypes[BOTTOM] = jBufferType;
    c->bufferTypes[TOP]    = jBufferType;

    c->sdispls[LEFT]   = ((imaxLocal + 2) + 1) * dblsize;             // send left
    c->sdispls[RIGHT]  = ((imaxLocal + 2) + imaxLocal) * dblsize;     // send right
    c->sdispls[BOTTOM] = ((imaxLocal + 2) + 1) * dblsize;             // send bottom
    c->sdispls[TOP]    = (jmaxLocal * (imaxLocal + 2) + 1) * dblsize; // send top

    c->rdispls[LEFT]   = (imaxLocal + 2) * dblsize;                         // recv left
    c->rdispls[RIGHT]  = ((imaxLocal + 2) + (imaxLocal + 1)) * dblsize;     // recv right
    c->rdispls[BOTTOM] = 1 * dblsize;                                       // recv bottom
    c->rdispls[TOP]    = ((jmaxLocal + 1) * (imaxLocal + 2) + 1) * dblsize; // recv top
#else
    c->imaxLocal = imax;
    c->jmaxLocal = jmax;
#endif
}
